package org.artifactory.repo.system;

import lombok.extern.slf4j.Slf4j;
import org.artifactory.api.config.CentralConfigService;
import org.artifactory.descriptor.config.MutableCentralConfigDescriptor;
import org.artifactory.repo.system.SystemRepos.SystemRepoConfig;
import org.artifactory.repo.system.suppliers.BuildInfoRepo;
import org.artifactory.repo.system.suppliers.EdgeRepo;
import org.artifactory.repo.system.suppliers.InTransitRepo;
import org.artifactory.repo.system.suppliers.LogShippingRepo;
import org.artifactory.repo.system.suppliers.PipeinfoRepo;
import org.artifactory.repo.system.suppliers.ReleaseBundleRepo;
import org.artifactory.repo.system.suppliers.SystemRepoBase;
import org.artifactory.repo.system.suppliers.UsageRepo;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.context.annotation.Lazy;
import org.springframework.stereotype.Component;

import java.util.Map;

import static org.artifactory.repo.system.SystemRepos.SystemRepoConfig.BUILDINFO;
import static org.artifactory.repo.system.SystemRepos.SystemRepoConfig.EDGE_UPLOADS;
import static org.artifactory.repo.system.SystemRepos.SystemRepoConfig.IN_TRANSIT;
import static org.artifactory.repo.system.SystemRepos.SystemRepoConfig.LOG_SHIPPING;
import static org.artifactory.repo.system.SystemRepos.SystemRepoConfig.PIPEINFO;
import static org.artifactory.repo.system.SystemRepos.SystemRepoConfig.RELEASE_BUNDLE;
import static org.artifactory.repo.system.SystemRepos.SystemRepoConfig.USAGE;

/**
 * This class will be called when possible when creating a system repo.
 * The map contains the system repos that can be built through the factory and their supplying classes
 * Note that not all system repos are actually built through here due to limitations such as only getting initialised
 * on startup or not needing to be saved to descriptor.
 *
 * @author Tomer Slutsky
 */
@Slf4j
@Lazy(true)
@Component
public class SystemRepoFactory {
    private final CentralConfigService centralConfigService;
    private final Map<SystemRepoConfig, SystemRepoBase> systemRepoSuppliers;

    @Autowired
    public SystemRepoFactory(CentralConfigService centralConfigService) {
        this.centralConfigService = centralConfigService;
        systemRepoSuppliers = Map.of(
                IN_TRANSIT,
                new InTransitRepo(centralConfigService),
                EDGE_UPLOADS,
                new EdgeRepo(centralConfigService),
                RELEASE_BUNDLE,
                new ReleaseBundleRepo(centralConfigService),
                PIPEINFO,
                new PipeinfoRepo(centralConfigService),
                BUILDINFO,
                new BuildInfoRepo(centralConfigService),
                LOG_SHIPPING,
                new LogShippingRepo(centralConfigService),
                USAGE,
                new UsageRepo(centralConfigService)
        );
    }

    public void buildSystemRepo(SystemRepoConfig repoConfig, String repoKey) {
        SystemRepoBase repo = systemRepoSuppliers.get(repoConfig);
        if (repo == null) {
            log.warn("Tried to build a system repo with a non existing system repo - {}", repoConfig.getRepoKey());
            return;
        }
        log.debug("Fetched system repo - {}", repoConfig.getRepoKey());
        MutableCentralConfigDescriptor mutableDescriptor = centralConfigService.getMutableDescriptor();
        repo.addSystemRepo(mutableDescriptor, repoKey);
    }

    public void buildSystemRepo(SystemRepoConfig repoConfig, String repoKey,
            MutableCentralConfigDescriptor mutableDescriptor) {
        SystemRepoBase repo = systemRepoSuppliers.get(repoConfig);
        if (repo == null) {
            log.warn("Tried to build a system repo with a non existing system repo - {}", repoConfig.getRepoKey());
            return;
        }
        log.debug("Fetched system repo - {}", repoConfig.getRepoKey());
        repo.addSystemRepo(mutableDescriptor, repoKey);
    }
}
