package org.artifactory.storage.db.init;

import org.artifactory.common.config.db.ArtifactoryDbProperties;
import org.artifactory.test.ArtifactoryHomeBoundTest;
import org.jfrog.storage.DbType;
import org.jfrog.storage.HikariJDBCDataSource;
import org.mockito.Mock;
import org.mockito.MockitoAnnotations;
import org.testng.annotations.BeforeMethod;
import org.testng.annotations.Test;

import java.sql.SQLException;

import static org.artifactory.common.config.db.ArtifactoryDbProperties.DEFAULT_POOL_TYPE;
import static org.mockito.ArgumentMatchers.anyString;
import static org.mockito.ArgumentMatchers.eq;
import static org.mockito.Mockito.times;
import static org.mockito.Mockito.verify;
import static org.mockito.Mockito.when;
import static org.testng.Assert.assertEquals;

@Test
public class CommonDataSourceFactoryTest extends ArtifactoryHomeBoundTest {

    @Mock
    ArtifactoryDbProperties artifactoryDbProperties;

    @BeforeMethod
    public void setUp() {
        MockitoAnnotations.initMocks(this);
        when(artifactoryDbProperties.getProperty(eq("connectivityCheckTimeoutMs"), anyString()))
                .thenReturn("5000");
    }

    @Test
    public void artifactoryTestInitTomcatDataSourceParameters() throws SQLException {
        prepareDefaultTomcatDataSource();

        ArtifactoryTomcatDataSource artifactoryTomcatDataSource = CommonDataSourceFactory
                .createDataSource(artifactoryDbProperties, true).unwrap(ArtifactoryTomcatDataSource.class);

        // When temporaryDbChannel channel set to true, it overwrites the default values.
        assertEquals(artifactoryTomcatDataSource.getMaxActive(), 1);
        assertEquals(artifactoryTomcatDataSource.getMaxIdle(), 1);
        assertEquals(artifactoryTomcatDataSource.getMinIdle(), 0);
    }

    @Test
    public void artifactoryTestInitArtifactoryDataSourceParameters() throws SQLException {
        prepareDefaultTomcatDataSource();

        ArtifactoryDataSource artifactoryDataSource = CommonDataSourceFactory
                .createDataSource(artifactoryDbProperties, true).unwrap(ArtifactoryDataSource.class);

        // When temporaryDbChannel channel set to true, it overwrites the default values.
        assertEquals(artifactoryDataSource.getMaxActive(), 1);
        assertEquals(artifactoryDataSource.getMaxIdle(), 1);
        assertEquals(artifactoryDataSource.getMinIdle(), 0);
    }

    @Test
    public void artifactoryTestTomcatDataSourceParameters() throws SQLException {
        prepareDefaultTomcatDataSource();

        ArtifactoryTomcatDataSource artifactoryTomcatDataSource = CommonDataSourceFactory
                .createDataSource(artifactoryDbProperties, false).unwrap(ArtifactoryTomcatDataSource.class);

        assertEquals(artifactoryTomcatDataSource.getMaxActive(), 5);
        assertEquals(artifactoryTomcatDataSource.getMaxIdle(), 5);
        assertEquals(artifactoryTomcatDataSource.getMinIdle(), 5);
    }

    @Test
    public void artifactoryTestInitHikariDataSourceParameters() throws SQLException {
        prepareDefaultHikariDataSource();

        ArtifactoryHikariDataSource artifactoryHikariDataSource = CommonDataSourceFactory
                .createDataSource(artifactoryDbProperties, true).unwrap(ArtifactoryHikariDataSource.class);
        // When temporaryDbChannel channel set to true, it overwrites the default values.
        assertEquals(artifactoryHikariDataSource.getMaximumPoolSize(), 1);
        assertEquals(artifactoryHikariDataSource.getMinimumIdle(), 0);
    }

    @Test
    public void artifactoryTestHikariDataSourceParameters() throws SQLException {
        prepareDefaultHikariDataSource();
        when(artifactoryDbProperties.getIntProperty("minIdle", 1)).thenReturn(1);

        ArtifactoryHikariDataSource artifactoryHikariDataSource = CommonDataSourceFactory
                .createDataSource(artifactoryDbProperties, false).unwrap(ArtifactoryHikariDataSource.class);

        assertEquals(artifactoryHikariDataSource.getMaximumPoolSize(), 5);
        assertEquals(artifactoryHikariDataSource.getMinimumIdle(), 1);
    }

    @Test
    public void artifactoryTestTomcatCreateUniqueDataSourceByType() throws SQLException {
        prepareDefaultTomcatDataSource();

        ArtifactoryTomcatDataSource artifactoryTomcatDataSource = CommonDataSourceFactory
                .createUniqueDataSourceByType(artifactoryDbProperties).unwrap(ArtifactoryTomcatDataSource.class);

        // The createUniqueDataSourceByType overwrites the default values.
        assertEquals(artifactoryTomcatDataSource.getMaxActive(), 1);
        assertEquals(artifactoryTomcatDataSource.getMaxIdle(), 1);
        assertEquals(artifactoryTomcatDataSource.getMinIdle(), 0);
        assertEquals(artifactoryTomcatDataSource.getInitialSize(), 0);
    }

    @Test
    public void artifactoryTestHikariCreateUniqueDataSourceByType() throws SQLException {
        prepareDefaultHikariDataSource();
        when(artifactoryDbProperties.getIntProperty("minIdle", 0)).thenReturn(1);
        when(artifactoryDbProperties.getLongProperty("leakDetectionThreshold", 60_000)).thenReturn(60_000L);

        HikariJDBCDataSource artifactoryHikariDataSource = CommonDataSourceFactory
                .createUniqueDataSourceByType(artifactoryDbProperties).unwrap(HikariJDBCDataSource.class);

        // The createUniqueDataSourceByType overwrites the default values.
        assertEquals(artifactoryHikariDataSource.getMaximumPoolSize(), 1);
        assertEquals(artifactoryHikariDataSource.getMinimumIdle(), 1);
        assertEquals(artifactoryHikariDataSource.getLeakDetectionThreshold(), 60_000);
    }

    @Test
    public void artifactoryTestTomcatWithDbLock() throws SQLException {
        prepareDefaultTomcatDataSource();

        ArtifactoryTomcatDataSource artifactoryTomcatDataSource = CommonDataSourceFactory
                .createDifferentDatasourceForDbLock(artifactoryDbProperties, true)
                .unwrap(ArtifactoryTomcatDataSource.class);

        // When temporaryDbChannel channel set to true, it overwrites the default values.
        assertEquals(artifactoryTomcatDataSource.getMaxActive(), 1);
        assertEquals(artifactoryTomcatDataSource.getMaxIdle(), 1);
        assertEquals(artifactoryTomcatDataSource.getMinIdle(), 0);
        verify(artifactoryDbProperties, times(0)).getLockingDbDriverClass();

        when(artifactoryDbProperties.getLockingDbSpecificType()).thenReturn(java.util.Optional.of(DbType.DERBY));
        artifactoryTomcatDataSource = CommonDataSourceFactory
                .createDifferentDatasourceForDbLock(artifactoryDbProperties, true)
                .unwrap(ArtifactoryTomcatDataSource.class);

        assertEquals(artifactoryTomcatDataSource.getMaxActive(), 1);
        assertEquals(artifactoryTomcatDataSource.getMaxIdle(), 1);
        assertEquals(artifactoryTomcatDataSource.getMinIdle(), 0);
        verify(artifactoryDbProperties, times(1)).getLockingDbDriverClass();
    }

    @Test
    public void artifactoryTestHikariWithDbLock() throws SQLException {
        prepareDefaultHikariDataSource();
        when(artifactoryDbProperties.getIntProperty("minIdle", 0)).thenReturn(1);
        when(artifactoryDbProperties.getLongProperty("leakDetectionThreshold", 60_000)).thenReturn(60_000L);
        when(artifactoryDbProperties.getLockingDbDriverClass()).thenReturn("org.apache.derby.jdbc.EmbeddedDriver");

        HikariJDBCDataSource artifactoryHikariDataSource = CommonDataSourceFactory
                .createDifferentDatasourceForDbLock(artifactoryDbProperties, true).unwrap(HikariJDBCDataSource.class);
        // When temporaryDbChannel channel set to true, it overwrites the default values.
        assertEquals(artifactoryHikariDataSource.getMaximumPoolSize(), 1);
        assertEquals(artifactoryHikariDataSource.getMinimumIdle(), 0);
        verify(artifactoryDbProperties, times(0)).getLockingDbDriverClass();

        when(artifactoryDbProperties.getLockingDbSpecificType()).thenReturn(java.util.Optional.of(DbType.DERBY));
        artifactoryHikariDataSource = CommonDataSourceFactory
                .createDifferentDatasourceForDbLock(artifactoryDbProperties, true).unwrap(HikariJDBCDataSource.class);
        assertEquals(artifactoryHikariDataSource.getMaximumPoolSize(), 1);
        assertEquals(artifactoryHikariDataSource.getMinimumIdle(), 0);
        verify(artifactoryDbProperties, times(1)).getLockingDbDriverClass();
    }

    // Prepares default params for tomcat data source.
    private void prepareDefaultTomcatDataSource() {
        when(artifactoryDbProperties.getProperty(ArtifactoryDbProperties.Key.POOL_TYPE, DEFAULT_POOL_TYPE))
                .thenReturn("tomcat-jdbc");
        when(artifactoryDbProperties.getMaxIdleConnections()).thenReturn(5);
        when(artifactoryDbProperties.getIntProperty("minIdle", 1)).thenReturn(5);
        when(artifactoryDbProperties.getMaxActiveConnections()).thenReturn(5);
        when(artifactoryDbProperties.getDbType()).thenReturn(DbType.DERBY);
    }

    // Prepares default params for hikari data source.
    // Params such as validationQueryTimeout or org.apache.derby.jdbc.EmbeddedDriver
    // are required to init the data source
    private void prepareDefaultHikariDataSource() {
        when(artifactoryDbProperties.getProperty(ArtifactoryDbProperties.Key.POOL_TYPE, DEFAULT_POOL_TYPE))
                .thenReturn("hikari");
        when(artifactoryDbProperties.getMaxActiveConnections()).thenReturn(5);
        when(artifactoryDbProperties.getDbType()).thenReturn(DbType.DERBY);
        when(artifactoryDbProperties.getDriverClass()).thenReturn("org.apache.derby.jdbc.EmbeddedDriver");
        when(artifactoryDbProperties.getIntProperty("validationQueryTimeout", 5000)).thenReturn(5000);
    }
}