/*
 *
 * Artifactory is a binaries repository manager.
 * Copyright (C) 2018 JFrog Ltd.
 *
 * Artifactory is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 * Artifactory is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with Artifactory.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package org.artifactory.storage.db.upgrades.common;

import ch.qos.logback.classic.util.ContextInitializer;
import org.apache.commons.io.IOUtils;
import org.artifactory.common.ConstantValues;
import org.artifactory.common.config.db.ArtifactoryDbProperties;
import org.artifactory.common.home.ArtifactoryHome;
import org.artifactory.storage.binstore.service.GarbageCollectionDatastoreService;
import org.artifactory.storage.db.LockingDBConverterAdapter;
import org.artifactory.storage.db.home.ArtifactoryUniqueFolderHomeBoundTest;
import org.artifactory.storage.db.init.ArtifactoryDataSource;
import org.artifactory.storage.db.init.CommonDataSourceFactory;
import org.artifactory.storage.db.itest.DbTestUtils;
import org.artifactory.storage.db.itest.UniqueJvmForkTestIdentifier;
import org.artifactory.storage.db.upgrades.external.itest.ExternalConversionScriptTest;
import org.artifactory.storage.db.util.ArtifactoryDbVersionUtil;
import org.artifactory.storage.db.version.ArtifactoryDBVersion;
import org.artifactory.version.ArtifactoryVersionProvider;
import org.artifactory.version.CompoundVersionDetails;
import org.jfrog.common.ResourceUtils;
import org.jfrog.converter.DbConverterExecutionException;
import org.jfrog.converter.DbConverterRuntimeException;
import org.jfrog.converter.DbDetails;
import org.jfrog.converter.DbVersion;
import org.jfrog.converter.model.ConverterExecutionType;
import org.jfrog.converter.util.DbConverterUtil;
import org.jfrog.storage.DbType;
import org.jfrog.storage.JFrogDataSource;
import org.jfrog.storage.JdbcHelper;
import org.jfrog.storage.metric.MetricsDataSource;
import org.jfrog.storage.util.DbStatementUtils;
import org.jfrog.storage.util.DbUtils;
import org.mockito.Mockito;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Qualifier;
import org.springframework.test.context.ContextConfiguration;
import org.springframework.test.context.testng.AbstractTestNGSpringContextTests;
import org.testng.annotations.AfterClass;
import org.testng.annotations.AfterMethod;
import org.testng.annotations.BeforeClass;
import org.testng.annotations.BeforeMethod;
import org.testng.annotations.Test;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.net.URL;
import java.sql.SQLException;
import java.util.List;

import static org.artifactory.storage.db.itest.DbTestUtils.initAndCreateUniqueDbNamespaceIfNeeded;
import static org.jfrog.common.FileUtils.writeFile;
import static org.jfrog.storage.util.DbUtils.doStreamWithConnection;
import static org.jfrog.storage.util.DbUtils.withConnection;
import static org.testng.Assert.assertEquals;

/**
 * Date: 8/5/13 9:58 AM
 *
 * @author freds
 */
@Test(groups = "dbtest-upgrade")
@ContextConfiguration(locations = {"classpath:spring/db-upgrade-test-context.xml"})
public abstract class UpgradeBaseTest extends AbstractTestNGSpringContextTests {
    protected static final String RESOURCE_FOR_SUBSTITUTION_NAME = "_v18_ui_session.sql";

    static {
        // use the itest logback config
        URL url = UpgradeBaseTest.class.getClassLoader().getResource("logback-dbtest.xml");
        System.setProperty(ContextInitializer.CONFIG_FILE_PROPERTY, url.getPath());
    }

    @Autowired
    protected JdbcHelper jdbcHelper;

    @Autowired
    @Qualifier("dbProperties")
    protected ArtifactoryDbProperties dbProperties;

    @Autowired
    protected GarbageCollectionDatastoreService garbageCollectionDatastoreService;

    private ArtifactoryUniqueFolderHomeBoundTest artifactoryHomeBoundTest;

    private static InputStream getPreviousDbSchemaSql(String previousVersion, DbType dbType) {
        String dbConfigDir = DbUtils.getDbTypeNameForSqlResources(dbType);
        return ResourceUtils.getResource(
                "/upgrades/" + previousVersion + "/ddl/" + dbConfigDir + "/" + dbConfigDir + ".sql");
    }

    private static InputStream getPreviousImportSql(String previousVersion, String dataFileName) {
        return ResourceUtils.getResource(
                "/upgrades/" + previousVersion + "/data/" + dataFileName + ".sql");
    }

    @BeforeClass
    @Override
    protected void springTestContextPrepareTestInstance() throws Exception {
        artifactoryHomeBoundTest = createArtifactoryHomeTest();
        artifactoryHomeBoundTest.bindArtifactoryHome();
        initAndCreateUniqueDbNamespaceIfNeeded();
        super.springTestContextPrepareTestInstance();
        contextCreated();
        rollBackToInitialState();
        runConvertersUpTo(upToVersion());
        convertLockingDBIfPresent();
    }

    private void runConvertersUpTo(ArtifactoryDBVersion upToVersion) {
        final List<DbVersion> dbVersions =
                ArtifactoryDbVersionUtil.versionsBetween(ArtifactoryDBVersion.v10, upToVersion);
        try {
            DbConverterUtil.calculateAndExecute(new DbDetails(dbProperties.getDbType(), jdbcHelper),
                    new DbVersion[] {},
                    dbVersions.toArray(DbVersion[]::new), upToVersion.systemVersion(),
                    ArtifactoryHome.get().getServerId());
        } catch (DbConverterExecutionException | SQLException e) {
            throw new DbConverterRuntimeException(e.getMessage(), e);
        }
    }

    private void runConvertersUpTo(ArtifactoryVersionProvider skipUntilVersion, ArtifactoryDBVersion upToVersion) {
        final List<DbVersion> dbVersions =
                ArtifactoryDbVersionUtil.versionsBetween(ArtifactoryDBVersion.v10, upToVersion);
        try {
            final CompoundVersionDetails compoundVersionDetails = Mockito.mock(CompoundVersionDetails.class);
            Mockito.when(compoundVersionDetails.getVersion()).thenReturn(skipUntilVersion.get());
            DbConverterUtil.calculateAndExecute(new DbDetails(dbProperties.getDbType(), jdbcHelper),
                    ArtifactoryDBVersion.getAlreadyApplied(compoundVersionDetails, ConverterExecutionType.SEQ_SYNC),
                    dbVersions.toArray(DbVersion[]::new), upToVersion.systemVersion(),
                    ArtifactoryHome.get().getServerId());
        } catch (DbConverterExecutionException | SQLException e) {
            throw new DbConverterRuntimeException(e.getMessage(), e);
        }
    }

    protected void contextCreated() throws Exception {
        // NO OP
    }

    /**
     * if LockingDB is present: Set up LockingDBConverterAdapter and convert
     *
     * @throws Exception
     */
    private void convertLockingDBIfPresent() throws Exception {
        if (dbProperties.getDbType().equals(DbType.POSTGRESQL) &&
                dbProperties.getLockingDbSpecificType().isPresent() &&
                dbProperties.getLockingDbSpecificType().get().equals(DbType.POSTGRESQL)) {
            JFrogDataSource lockingDBDataSource =
                    CommonDataSourceFactory.createDifferentDatasourceForDbLock(dbProperties, true);
            JdbcHelper lockingDBJdbcHelper = new JdbcHelper(lockingDBDataSource);
            DbUtils.doWithConnection(lockingDBJdbcHelper,
                    conn -> DbTestUtils.dropExistingObjects(conn, dbProperties.getLockingDbSpecificType().get()));
            executeSqlStream(getPreviousDbSchemaSql("v550c", dbProperties.getDbType()), lockingDBJdbcHelper);
            LockingDBConverterAdapter lockingDBConverterAdapter =
                    new LockingDBConverterAdapter(ArtifactoryHome.get(),
                            new DbDetails(dbProperties.getLockingDbSpecificType().get(),
                                    lockingDBJdbcHelper));
            lockingDBConverterAdapter.convert(null, null);
        }
    }

    /**
     * Sets up an external conversion script that overrides one (arbitrarily picked) built-in conversion file during
     * conversion, with an externally provided conversion script. This is a bit hacky but required because
     * {@link this#springTestContextPrepareTestInstance} runs before any test can intervene and is invoking the entire
     * conversion chain. This is called only once, and explicitly at
     * {@link ExternalConversionScriptTest}
     *
     * @param emptyScript - indicates whether the script should be empty of content
     */
    protected void prepareTestInstanceForExternalConversion(boolean emptyScript) throws Exception {
        // We are already ArtifactoryHome bound. Rollback, prepare the external script, and convert again
        rollBackToInitialState();
        ArtifactoryHome.get().getArtifactoryProperties()
                .setProperty(ConstantValues.allowExternalConversionScripts.getPropertyName(), "true");
        setupExternalConversionScript(emptyScript);
        runConvertersUpTo(upToVersion());
    }

    private void setupExternalConversionScript(boolean emptyScript) {
        String externalConversionScript = "";
        File conversionsDir = new File(ArtifactoryHome.get().getEtcDir() + "/conversion");
        String sqlResource = "external_conversion_script.sql";
        String dbType = DbUtils.getDbTypeNameForSqlResources(dbProperties.getDbType());
        if (dbType.equals("oracle") || dbType.equals("mssql")) {
            sqlResource = "oracle_" + sqlResource;
        }
        conversionsDir.mkdir();
        if (!emptyScript) {
            String scriptResourcePath = "/sql/" + sqlResource;
            externalConversionScript = ResourceUtils.getResourceAsString(scriptResourcePath);
        }
        writeFile(conversionsDir + "/" + dbType + RESOURCE_FOR_SUBSTITUTION_NAME, externalConversionScript.getBytes());
    }

    private ArtifactoryUniqueFolderHomeBoundTest createArtifactoryHomeTest() {
        return new ArtifactoryUniqueFolderHomeBoundTest(UniqueJvmForkTestIdentifier.get());
    }

    @AfterClass
    protected void verifyDbResourcesReleased() throws SQLException {
        ArtifactoryDataSource ds;
        JFrogDataSource jFrogDataSource = jdbcHelper.getDataSource();
        // make sure there are no active connections
        if (jFrogDataSource.isWrapperFor(MetricsDataSource.class)) {
            ds = jFrogDataSource.unwrap(MetricsDataSource.class).unwrap(ArtifactoryDataSource.class);
        } else {
            ds = jFrogDataSource.unwrap(ArtifactoryDataSource.class);
        }
        assertEquals(ds.getActiveConnectionsCount(), 0, "Found " + ds.getActiveConnectionsCount() +
                " active connections after test ended");
        artifactoryHomeBoundTest.unbindArtifactoryHome();
    }

    @BeforeMethod
    public void bindArtifactoryHome() {
        artifactoryHomeBoundTest.bindArtifactoryHome();
    }

    @AfterMethod
    public void unbindArtifactoryHome() {
        artifactoryHomeBoundTest.unbindArtifactoryHome();
    }

    @AfterClass
    public void cleanEnv() {
        artifactoryHomeBoundTest.cleanEnv();
    }

    protected void rollBackToInitialState() throws IOException, SQLException {
        // Making DB looks like v10 with data
        DbUtils.doWithConnection(jdbcHelper, conn -> DbTestUtils.dropExistingObjects(conn, dbProperties.getDbType()));
        executeSqlStream(getPreviousDbSchemaSql("v10", dbProperties.getDbType()));
        executeSqlStream(getPreviousImportSql("v10", "user-group"));
        executeSqlStream(getPreviousImportSql("v10", "acls"));
        executeSqlStream(getPreviousImportSql("v10", "builds"));
        executeSqlStream(getPreviousImportSql("v10", "nodes"));
    }

    protected void resetToVersion(ArtifactoryDBVersion resetTo) throws IOException, SQLException {
        rollBackToInitialState();
        runConvertersUpTo(resetTo);
    }

    protected void resetToVersion(ArtifactoryVersionProvider skipUntilVersion, ArtifactoryDBVersion resetTo)
            throws IOException, SQLException {
        rollBackToInitialState();
        runConvertersUpTo(skipUntilVersion, resetTo);
    }

    protected void importSql(String resourcePath) {
        try (InputStream resource = ResourceUtils.getResource(resourcePath)) {
            executeSqlStream(resource);
        } catch (Exception e) {
            throw new RuntimeException(e);
        }
    }

    protected boolean columnExists(String table, String column) throws SQLException {
        return withConnection(jdbcHelper,
                conn -> DbTestUtils.columnExists(conn, dbProperties.getDbType(), table, column));
    }

    protected boolean tableExists(String table) throws SQLException {
        return withConnection(jdbcHelper, conn -> DbTestUtils.tableExists(conn, dbProperties.getDbType(), table));
    }

    protected int columnSize(String table, String columnName) throws SQLException {
        return withConnection(jdbcHelper,
                conn -> DbTestUtils.getColumnSize(conn, dbProperties.getDbType(), table, columnName));
    }

    protected boolean isColumnNullable(String table, String columnName) throws SQLException {
        return withConnection(jdbcHelper,
                conn -> DbUtils.isColumnNullable(conn.getMetaData(), dbProperties.getDbType(), table, columnName));
    }

    protected void executeSqlStream(InputStream in) throws IOException, SQLException {
        executeSqlStream(in, jdbcHelper);
    }

    protected void executeSqlStream(InputStream in, JdbcHelper jdbcHelper) throws IOException, SQLException {
        try {
            doStreamWithConnection(jdbcHelper, conn -> DbStatementUtils.executeSqlStream(conn, in));
        } finally {
            IOUtils.closeQuietly(in);
        }
    }

    /**
     * Override this method in child classed to set version up to which converters should be executed.
     */

    protected ArtifactoryDBVersion upToVersion() {
        throw new UnsupportedOperationException("Set version up to which converters should be applied.");
    }
}
